"""  Module for Job Shop Scheduling Problem (JSSP)."""
from collections import defaultdict
from typing import Literal
try:
    from qlauncher.problems.problem_formulations.jssp.qiskit_scheduler import get_jss_hamiltonian
except ModuleNotFoundError:
    pass
from qlauncher.base import Problem


class JSSP(Problem):
    """
    Class for Job Shop Scheduling Problem.

    This class represents Job Shop Scheduling Problem (JSSP) which is a combinatorial optimization problem that involves 
    scheduling a set of jobs on a set of machines. Each job consists of a sequence of operations that must be performed 
    on different machines. The objective is to find a schedule that minimizes the makespan, i.e., the total time required
    to complete all jobs. The class contains an instance of the problem, so it can be passed into QLauncher.


    Attributes:
        max_time (int): The maximum time for the scheduling problem.
        onehot (str): The one-hot encoding method to be used.
        optimization_problem (bool): Flag indicating whether the problem is an optimization problem or a decision problem.
        results (dict): Dictionary to store the results of the problem instance.

    """

    def __init__(self, max_time: int, instance: dict[str, list[tuple[str, int]]],
                 instance_name: str = 'unnamed', optimization_problem: bool = False, onehot: Literal['exact', 'quadratic'] = 'exact') -> None:
        super().__init__(instance=instance, instance_name=instance_name)
        self.max_time = max_time
        self.onehot = onehot
        self.optimization_problem = optimization_problem

        self.h_d, self.h_o, self.h_pos_by_label, self.h_label_by_pos = get_jss_hamiltonian(self.instance, max_time,
                                                                                           onehot)

        self.results = {'instance_name': instance_name,
                        'max_time': max_time,
                        'onehot': onehot,
                        'H_pos_by_label': self.h_pos_by_label,
                        'H_label_by_pos': self.h_label_by_pos}

        self.variant = 'optimization' if optimization_problem else 'decision'

    @property
    def setup(self) -> dict:
        return {
            'max_time': self.max_time,
            'onehot': self.onehot,
            'optimization_problem': self.optimization_problem,
            'instance_name': self.instance_name
        }

    def _get_path(self) -> str:
        return f'{self.name}@{self.instance_name}@{self.max_time}@{"optimization" if self.optimization_problem else "decision"}@{self.onehot}'

    @staticmethod
    def from_preset(instance_name: str, **kwargs) -> "JSSP":
        match instance_name:
            case 'default':
                max_time = 3
                instance = {"cupcakes": [("mixer", 2), ("oven", 1)],
                            "smoothie": [("mixer", 1)],
                            "lasagna": [("oven", 2)]}
            case _:
                raise ValueError(f"Instance {instance_name} does not exist choose instance_name from the following: ('toy')")

        return JSSP(max_time=max_time, instance=instance, instance_name=instance_name, **kwargs)

    @classmethod
    def from_file(cls, path: str, **kwargs) -> "JSSP":
        job_dict = defaultdict(list)
        with open(path, 'r', encoding='utf-8') as file_:
            file_.readline()
            for i, line in enumerate(file_):
                lint = list(map(int, line.split()))
                job_dict[i + 1] = [x for x in
                                   zip(lint[::2],  # machines
                                       lint[1::2]  # operation lengths
                                       )]
        return JSSP(instance=job_dict, **kwargs)



from typing import Literal

import numpy as np
import networkx as nx

from qlauncher.problems.problem_initialization import TSP

from qlauncher.hampy import Equation
import qlauncher.hampy as hampy
from qlauncher.hampy.utils import shift_affected_qubits


def make_non_collision_hamiltonian(node_count: int, quadratic=False):
    """
    Creates a Hamiltonian representing constraints for the TSP problem. (Each node visited only once, one node per timestep)
    Qubit mapping: [step1:[node1, node2,... noden]],[step2],...[stepn]

    Args:
        node_count: Number of nodes in the TSP problem
        quadratic: Whether to encode as a QUBO problem

    Returns:
        np.ndarray: Hamiltonian representing the constraints
    """

    n = node_count**2
    eq = Equation(n)

    # hampy.one_in_n() takes a long time to execute with large qubit counts.
    # We can exploit the nature of tsp constraints and create the operator for the first timestep and then shift it for the rest of the timesteps
    # Same with nodes below.
    # I'm pretty sure this works as intended...

    # Ensure that at each timestep only one node is visited
    t0_op = hampy.one_in_n(
        [i for i in range(node_count)],
        eq.size,
        quadratic=quadratic
    )

    for timestep in range(node_count):
        shift = shift_affected_qubits(t0_op, timestep * node_count)
        eq += shift

    # Ensure that each node is visited only once
    n0_op = hampy.one_in_n(
        [timestep * node_count for timestep in range(node_count)],
        eq.size,
        quadratic=quadratic
    )

    for node in range(node_count):
        shift = shift_affected_qubits(n0_op, node)
        eq += shift

    return -1 * eq.hamiltonian


def make_connection_hamiltonian(edge_costs: np.ndarray, return_to_start: bool = True) -> np.ndarray:
    """
    Creates a Hamiltonian that represents the costs of picking each path.

    Args:
        tsp_matrix: Edge cost matrix of the TSP problem

    Returns:
        np.ndarray: Optimal chain of nodes to visit
    """
    node_count = edge_costs.shape[0]
    if len(edge_costs.shape) != 2 or edge_costs.shape[1] != node_count:
        raise ValueError("edge_costs must be a square matrix")

    n = node_count**2
    eq = Equation(n)

    for timestep in range(node_count - 1):
        for node in range(node_count):
            for next_node in range(node_count):
                if node == next_node:
                    continue
                and_hamiltonian = (
                    eq[node + timestep * node_count]
                    & eq[next_node + (timestep + 1) * node_count]
                )
                eq += edge_costs[node, next_node] * and_hamiltonian

    if not return_to_start:
        return eq.hamiltonian

    # Add cost of returning to the first node
    for node in range(node_count):
        for node2 in range(node_count):
            and_hamiltonian = (
                eq[node + (node_count - 1) * node_count] & eq[node2]
            )
            eq += edge_costs[node, node2] * and_hamiltonian

    return eq.hamiltonian


def problem_to_hamiltonian(problem: TSP, constraints_weight: int = 5, costs_weight: int = 1, return_to_start: bool = True, onehot: Literal['exact', 'quadratic'] = 'exact') -> np.ndarray:
    """
    Creates a Hamiltonian for the TSP problem.

    Args:
        problem: TSP problem instance
        quadratic: Whether to encode as a quadratic Hamiltonian
        constraints_weight: Weight of the constraints in the Hamiltonian
        costs_weight: Weight of the costs in the Hamiltonian

    Returns:
        np.ndarray: Hamiltonian representing the TSP problem
    """
    instance_graph = problem.instance

    edge_costs = nx.to_numpy_array(instance_graph)
    # discourage breaking the constraints
    edge_costs += np.eye(len(edge_costs)) * np.max(edge_costs)
    scaled_edge_costs = edge_costs.astype(np.float32) / np.max(edge_costs)

    node_count = len(instance_graph.nodes)

    constraints = make_non_collision_hamiltonian(node_count, quadratic=(onehot == 'quadratic'))
    costs = make_connection_hamiltonian(scaled_edge_costs, return_to_start=return_to_start)

    return constraints * constraints_weight + costs * costs_weight



""" All problems together """
from qlauncher.base import Problem
from . import problem_formulations as _
from .problem_initialization import Raw, MaxCut, EC, QATM, JSSP, TSP, GraphColoring, TabularML, Molecule

__all__ = ['Problem', 'Raw', 'MaxCut', 'EC', 'QATM', 'JSSP', 'TSP', 'GraphColoring', 'TabularML', 'Molecule']



""" This module contains the Raw class."""
from typing import Any

import numpy as np
from qiskit.quantum_info import SparsePauliOp

from qlauncher.base import Problem, formatter


class Raw(Problem):
    """ Meta class for raw problem """
    __cached_classes__: dict[str, type] = {}

    @staticmethod
    def _cache_class(problem_type: str):
        if problem_type in Raw.__cached_classes__:
            return Raw.__cached_classes__[problem_type]

        cls = type(problem_type, (Raw, ), {})
        formatter(cls, problem_type)(_raw_formatter)
        Raw.__cached_classes__[problem_type] = cls
        return cls

    @staticmethod
    def _auto_map_problem(obj: Any) -> str:
        """Automatically maps obj into str.

        Args:
            obj (Any): obj of some problem class.

        Returns:
            str: name of problem class that obj is in.
        """
        if isinstance(obj, SparsePauliOp):
            return 'hamiltonian'
        if isinstance(obj, tuple) and len(obj) == 2 and \
                isinstance(obj[0], np.ndarray) and isinstance(obj[1], (int, float)):
            return 'qubo'
        try:
            from dimod.binary.binary_quadratic_model import BinaryQuadraticModel
            if isinstance(obj, BinaryQuadraticModel):
                return 'bqm'
        except ImportError:
            pass
        return 'none'

    def __new__(cls, instance: Any, instance_name: str = 'Raw'):
        if cls is not Raw:
            return super().__new__(cls)
        problem_type = cls._auto_map_problem(instance)
        true_cls = cls._cache_class(problem_type)
        return true_cls(instance, instance_name)


def _raw_formatter(raw: Raw) -> Any:
    return raw.instance



import ast
import numpy as np
from pyqubo import Spin
from qlauncher.base import adapter, formatter
from qlauncher.problems.problem_initialization import Raw


@adapter('qubo', 'bqm')
def qubo_to_bqm(qubo_with_offset) -> dict:
    qubo, offset = qubo_with_offset
    bqm, _ = QUBOMatrix(qubo, offset).qubo_matrix_into_bqm()
    return bqm


class QUBOMatrix:
    def __init__(self, qubo_matrix, offset):
        self.qubo_matrix = qubo_matrix
        self.offset = offset

        if type(self.qubo_matrix) == str:
            self.qubo_matrix = np.array(ast.literal_eval(self.qubo_matrix))

        self.symetric = True
        self._check_if_symetric()

    def _check_if_symetric(self):
        """
        Function to check if matrix is symetric
        """
        self.symetric = (self.qubo_matrix.transpose()
                         == self.qubo_matrix).all()
        if not self.symetric:
            self.qubo_matrix = self._remove_lower_triangle(self.qubo_matrix)

    def _remove_lower_triangle(self, matrix):
        """
        Function to remove lower triangle from matrix
        """
        for i in range(len(matrix)):
            for j in range(len(matrix)):
                if i > j:
                    matrix[i][j] = 0
        return matrix

    def _get_values_and_qubits(self, matrix):
        """
        Function to get values and qubits from matrix in form of dictionary
        where keys are indexes of qubits and values are values of matrix
        The function does not take into account zeros in matrix
        Example:
        matrix = [[0,1,2],[1,0,3],[2,3,0]]
        result = {(0,1):1, (0,2):2, (1,0):1, (1,2):3, (2,0):2, (2,1):3}
        Function also return second value which is the number of qubits
        """
        result = {
            (x, y): c for y, r in enumerate(matrix) for x, c in enumerate(r) if c != 0
        }
        return result, len(matrix)

    def qubo_matrix_into_bqm(self):
        values_and_qubits, number_of_qubits = self._get_values_and_qubits(
            self.qubo_matrix
        )
        qubits = [Spin(f"x{i}") for i in range(number_of_qubits)]
        H = 0
        for (x, y), value in values_and_qubits.items():
            if self.symetric:
                H += value / len(set((x, y))) * qubits[x] * qubits[y]
            else:
                H += value * qubits[x] * qubits[y]
        model = H.compile()
        bqm = model.to_bqm()
        bqm.offset += self.offset
        return bqm, model


@formatter(Raw, 'bqm')
def Rawbqm(problem: Raw):
    return problem.instance



"""
`equations` module provides additional binary operations for Hampy objects.

It's goal is too simplify the creation of more complex problem implementations, by creating them with use of smaller ones.
"""
from copy import copy
from qiskit.quantum_info import SparsePauliOp
from .object import Equation, Variable


def one_in_n(variables: list[int | Variable], size: int | None = None, quadratic: bool = False) -> Equation:
    """
    Generates Equation for One in N problem.

    One in N returns True if and only if exactly one of targeted indexes in 1, and all others are 0.

    Args:
        variables (list[int  |  Variable]): Triggered variables or variable indexes
        size (int | None, optional): Size of problem, if not given it takes the first found Variable.size value. Defaults to None.

    Returns:
        Equation: Equation returning True if exactly one of passed indexes is 1, False otherwise
    """
    if size is None:
        for var in variable:
            if isinstance(var, Variable):
                size = var.size
                break

    eq = Equation(size)
    new_variables = set()
    for var in copy(variables):
        if isinstance(var, int):
            new_variables.add(eq.get_variable(var))
        elif isinstance(var, Variable):
            new_variables.add(eq.get_variable(var.index))

    if quadratic:
        for variable in new_variables:
            eq += variable
        I = SparsePauliOp.from_sparse_list([('I', [], 1)], size)
        hamiltonian = I - eq.hamiltonian
        return Equation(hamiltonian.compose(hamiltonian))

    for variable in new_variables:
        equation = variable
        for different_var in new_variables - {variable}:
            equation &= ~different_var
        eq |= equation

    return eq



from abc import ABC, abstractmethod
from dataclasses import dataclass
import pickle
from typing import Any, Literal
from collections.abc import Callable
import logging


AVAILABLE_FORMATS = Literal['hamiltonian', 'qubo', 'bqm', 'none', 'fn', 'tabular_ml']


@dataclass
class Result:
    best_bitstring: str
    best_energy: float
    most_common_bitstring: str
    most_common_bitstring_energy: float
    distribution: dict
    energies: dict
    num_of_samples: int
    average_energy: float
    energy_std: float
    result: Any

    def __str__(self):
        return f"Result(bitstring={self.best_bitstring}, energy={self.best_energy})"

    def __repr__(self):
        return str(self)

    def best(self):
        return self.best_bitstring, self.best_energy

    def most_common(self):
        return self.most_common_bitstring, self.most_common_bitstring_energy

    @staticmethod
    def from_distributions(bitstring_distribution: dict[str, float], energy_distribution: dict[str, float], result: Any = None) -> "Result":
        """
        Constructs the Result object from Dictionary with bitstring to num of occurrences,
        dictionary mapping bitstring to energy and optional result (rest)
        """
        best_bitstring = min(energy_distribution, key=energy_distribution.get)
        best_energy = energy_distribution[best_bitstring]
        most_common_bitstring = max(bitstring_distribution, key=bitstring_distribution.get)
        most_common_bitstring_energy = energy_distribution[most_common_bitstring]
        num_of_samples = int(sum(bitstring_distribution.values()))

        mean_value = sum(energy_distribution[bitstring] * occ for bitstring, occ in bitstring_distribution.items()) / num_of_samples
        std = 0
        for bitstring, occ in bitstring_distribution.items():
            std += occ * ((energy_distribution[bitstring] - mean_value)**2)
        std = (std/(num_of_samples-1))**0.5
        return Result(
            best_bitstring,
            best_energy,
            most_common_bitstring,
            most_common_bitstring_energy,
            bitstring_distribution,
            energy_distribution,
            num_of_samples,
            mean_value,
            std,
            result
        )


class Backend:
    """
    Abstract class representing a backend for quantum computing.

    Attributes:
        name (str): The name of the backend.
        path (str | None): The path to the backend (optional).
        parameters (list): A list of parameters for the backend (optional).

    """

    def __init__(self, name: str, parameters: list | None = None) -> None:
        self.name: str = name
        self.is_device = name == 'device'
        self.path: str | None = None
        self.parameters = parameters if parameters is not None else []
        self.logger: logging.Logger | None = None

    def set_logger(self, logger: logging.Logger):
        self.logger = logger

    def _get_path(self):
        return f'{self.name}'


class Problem(ABC):
    """
    Abstract class for defining Problems.

    Attributes:
        variant (str): The variant of the problem. The default variant is "Optimization".
        path (str | None): The path to the problem.
        name (str): The name of the problem.
        instance_name (str): The name of the instance.
        instance (any): An instance of the problem.

    """

    _problem_id = None

    def __init__(self, instance: Any, instance_name: str = 'unnamed') -> None:
        """
        Initializes a Problem instance.

        Params:
            instance (any): An instance of the problem.
            instance_name (str | None): The name of the instance.

        Returns:
            None
        """
        self.instance: Any = instance
        self.instance_name = instance_name
        self.variant: str = 'Optimization'
        self.path: str | None = None
        self.name = self.__class__.__name__.lower()

    @classmethod
    def from_file(cls: type['Problem'], path: str) -> 'Problem':
        with open(path, 'rb') as f:
            instance = pickle.load(f)
        return cls(instance)

    @staticmethod
    def from_preset(instance_name: str, **kwargs):
        raise NotImplementedError()

    def __init_subclass__(cls) -> None:
        if Problem not in cls.__bases__:
            return
        cls._problem_id = cls

    def read_result(self, exp, log_path):
        """
        Reads a result from a file.

        Args:
            exp: The experiment.
            log_path: The path to the log file.

        Returns:
            The result.
        """
        exp += exp  # ?: this is perplexing
        with open(log_path, 'rb') as file:
            res = pickle.load(file)
        return res

    def analyze_result(self, result) -> Any:
        """
        Analyzes the result.

        Args:
            result: The result.

        """
        raise NotImplementedError()


class Algorithm(ABC):
    """
    Abstract class for Algorithms.

    Attributes:
        name (str): The name of the algorithm, derived from the class name in lowercase.
        path (str | None): The path to the algorithm, if applicable.
        parameters (list): A list of parameters for the algorithm.
        alg_kwargs (dict): Additional keyword arguments for the algorithm.

    Abstract methods:
        __init__(self, **alg_kwargs): Initializes the Algorithm object.
        _get_path(self) -> str: Returns the common path for the algorithm.
        run(self, problem: Problem, backend: Backend): Runs the algorithm on a specific problem using a backend.
    """
    _algorithm_format: AVAILABLE_FORMATS = 'none'

    def __init__(self, **alg_kwargs) -> None:
        self.name: str = self.__class__.__name__.lower()
        self.path: str | None = None
        self.parameters: list = []
        self.alg_kwargs = alg_kwargs

    def parse_result_to_json(self, o: object) -> dict:
        """Parses results so that they can be saved as a JSON file.

        Args:
            o (object): The result object to be parsed.

        Returns:
            dict: The parsed result as a dictionary.
        """
        print('Algorithm does not have the parse_result_to_json method implemented')
        return o.__dict__

    @abstractmethod
    def run(self, problem: Problem, backend: Backend, formatter: Callable) -> Result:
        """Runs the algorithm on a specific problem using a backend.

        Args:
            problem (Problem): The problem to be solved.
            backend (Backend): The backend to be used for execution.
        """



"""DWave backends"""

from abc import abstractmethod, ABC
from qlauncher.base import Backend


from qlauncher.exceptions import DependencyError

try:
    from dimod import Sampler
    from dwave.system import DWaveSampler, EmbeddingComposite
    from dwave.samplers import SimulatedAnnealingSampler, TabuSampler, SteepestDescentSampler
except ImportError as e:
    raise DependencyError(e, install_hint='dwave') from e


class BQMBackend(Backend, ABC):
    """Base dwave backend"""

    def __init__(self, name: str, parameters: list | None = None) -> None:
        super().__init__(name, parameters)
        self.sampler = self._get_sampler()

    @abstractmethod
    def _get_sampler(self) -> Sampler:
        """Get a dimod sampler"""


class TabuBackend(BQMBackend):
    """Tabu search simulator backend"""

    def __init__(self, parameters: list | None = None) -> None:
        super().__init__("TabuBackend", parameters)

    def _get_sampler(self):
        return TabuSampler()


class SimulatedAnnealingBackend(BQMBackend):
    """Simulated annealing simulator backend"""

    def __init__(self, parameters: list | None = None) -> None:
        super().__init__("SimulatedAnnealingBackend", parameters)

    def _get_sampler(self) -> Sampler:
        return SimulatedAnnealingSampler()


class SteepestDescentBackend(BQMBackend):
    """Steepest descent simulator backend"""

    def __init__(self, parameters: list | None = None) -> None:
        super().__init__("SteepestDescentBackend", parameters)

    def _get_sampler(self) -> Sampler:
        return SteepestDescentSampler()


class DwaveBackend(BQMBackend):
    """Real Dwave device backend"""

    def __init__(self, parameters: list | None = None) -> None:
        super().__init__("DwaveDevice", parameters)

    def _get_sampler(self) -> Sampler:
        return EmbeddingComposite(DWaveSampler())



"""Wrapper for QLauncher that enables the user to launch tasks asynchronously (futures + multiprocessing)"""
from collections.abc import Callable
from typing import Any
from threading import Event, Thread
import time
import weakref
import multiprocess

from pathos.multiprocessing import _ProcessPool

from qlauncher.base.base import Result


def get_timeout(max_timeout: int | float | None, start: int | float) -> float | None:
    """
    Get timeout to wait on an event, useful when awaiting multiple tasks and total timeout must be max_timeout.

    Args:
        max_timeout (int | float | None): Total allowed timeout, None = infinite wait.
        start (int | float): Await start timestamp (time.time())

    Returns:
        int | float | None: Remaining timeout or None if max_timeout was None
    """
    if max_timeout is None:
        return None
    return max_timeout - (time.time() - start)


class _InnerAQLTask:
    """
        Task object returned to user, so that dependencies can be created.

        Attributes:
            task (Callable): function that gets executed asynchronously
            dependencies (list[AQLTask]): Optional dependencies. The task will wait for all its dependencies to finish, before starting.
            callbacks (list[Callable]): Callbacks ran when the task finishes executing. 
                                        Task result is inserted as an argument to the function.
            pipe_dependencies (bool): If True results of tasks defined as dependencies will be passed as arguments to self.task. 
                                      Defaults to False.
    """

    def __init__(
        self,
        task: Callable,
        dependencies: list['AQLTask'] | None = None,
        callbacks: list[Callable] | None = None,
        pipe_dependencies: bool = False
    ) -> None:

        self.task = task
        self.dependencies = dependencies if dependencies is not None else []
        self.callbacks = callbacks if callbacks is not None else []
        self.pipe_dependencies = pipe_dependencies

        self._cancelled = False
        self._thread = None
        self._pool = _ProcessPool(processes=1)
        self._thread_made = Event()

        self._result = None
        self._done = False

    def _shutdown_subprocess(self):
        self._pool.close()
        self._pool.terminate()
        self._pool.join()

    def _async_task(self) -> Any:
        dep_results = [d.result() for d in self.dependencies]

        if self._cancelled:
            self._result = None
            self._done = True
            return

        res = self._pool.apply_async(self.task, args=(dep_results if self.pipe_dependencies else []))
        # Turns out you can't just outright kill threads (or futures) is so I have to do this, so that the thread knows to exit.
        while not self._cancelled:
            try:
                self._result = res.get(timeout=0.05)
                self._done = True
                return
            except multiprocess.context.TimeoutError:
                pass  # task not ready, check for cancel
            # For any other error originating from the task, shutdown and clean up subprocess then raise error again.
            except Exception as e:
                self._shutdown_subprocess()
                self._result = e
                self._done = True
                return

        if self._cancelled:
            self._shutdown_subprocess()  # kill res process

        self._result = res.get() if res.ready() else None
        self._done = True
        return

    def _target_task(self):
        # Main task + callbacks launch
        self._async_task()
        for cb in self.callbacks:
            cb(self._result)

    def _set_thread(self):
        self._thread = Thread(target=weakref.proxy(self)._target_task, daemon=True)  # set daemon so that thread quits as main process quits
        self._thread.start()
        self._thread_made.set()

    def start(self):
        """Start task execution."""
        if self._thread is not None or self._cancelled:
            raise ValueError("Cannot start, task already started or cancelled.")
        self._set_thread()

    def cancel(self) -> bool:
        """
        Attempt to cancel the task.

        Returns:
            bool: True if cancellation was successful
        """
        self._cancelled = True
        if self._thread is None:
            return True
        self._thread.join(0.1)
        return not self._thread.is_alive()

    def cancelled(self) -> bool:
        """
        Returns:
            bool: True if the task was cancelled by the user.
        """
        return self._cancelled

    def done(self) -> bool:
        """
        Returns:
            bool: True if the task had finished execution.
        """
        return self._done

    def running(self) -> bool:
        """
        Returns:
            bool: True if the task is currently executing.
        """
        if self._thread is None:
            return False
        return self._thread.is_alive()

    def result(self, timeout: float | int | None = None) -> Result | None:
        """
        Get result of running the task.
        Blocks the thread until task is finished.

        Args:
            timeout (float | int | None, optional): 
                    The maximum amount to wait for execution to finish.
                    If None, wait forever. If not None and time runs out, raises TimeoutError. 
                    Defaults to None.
        Returns:
            Result if future returned result or None when cancelled.
        """
        start = time.time()
        self._thread_made.wait(timeout=get_timeout(timeout, start))  # Wait until we start a thread
        self._thread.join(timeout=get_timeout(timeout, start))
        if self._thread.is_alive():
            self.cancel()
            raise TimeoutError  # thread still running after timeout
        if isinstance(self._result, BaseException):
            raise self._result
        return self._result


# Why like this? The inner task was not getting properly garbage collected when it was running,
# but this does and just cancels the inner task so it also gets garbage collected
# this is cursed :/
class AQLTask:
    """
        Task object returned to user, so that dependencies can be created.

        Attributes:
            task (Callable): function that gets executed asynchronously
            dependencies (list[AQLTask]): Optional dependencies. The task will wait for all its dependencies to finish, before starting.
            callbacks (list[Callable]): Callbacks ran when the task finishes executing. 
                                        Task result is inserted as an argument to the function.
            pipe_dependencies (bool): If True results of tasks defined as dependencies will be passed as arguments to self.task. 
                                      Defaults to False.
    """

    def __init__(
        self,
        task: Callable,
        dependencies: list['AQLTask'] | None = None,
        callbacks: list[Callable] | None = None,
        pipe_dependencies: bool = False
    ) -> None:
        self._inner_task = _InnerAQLTask(task, dependencies, callbacks, pipe_dependencies)
        weakref.finalize(self, self._inner_task.cancel)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._inner_task, name)



from collections.abc import Callable, Sequence
from typing import Any

from qiskit.circuit import QuantumCircuit, Parameter
from qiskit.primitives import BaseSampler, BaseSamplerV1, BaseSamplerV2
from qiskit.transpiler.passmanager import PassManager
from qiskit.providers import Options
import numpy as np

from qiskit_machine_learning.kernels.algorithms import QuantumKernelTrainer
from qiskit_machine_learning.kernels import TrainableFidelityQuantumKernel, FidelityQuantumKernel
from qiskit_machine_learning.state_fidelities import ComputeUncompute, BaseStateFidelity

from qlauncher.base.base import Backend, Problem, Algorithm, Result
from qlauncher.routines.qiskit.backends.qiskit_backend import QiskitBackend
from qlauncher.routines.cirq import CirqBackend


class ComputeUncomputeCustom(ComputeUncompute):
    """
        This is just :class:`qiskit_machine_learning.state_fidelities.ComputeUncompute` that checks 
        if a sampler is an instance of BaseSamplerV1 instead of BaseSampler.
        The reason was that classes basing BaseSampler were getting isinstance(cls(),BaseSampler) == False
        probably because of some qiskit shenanigans.
        """

    def __init__(
        self,
        sampler: BaseSampler | BaseSamplerV2,
        *,
        options: Options | None = None,
        local: bool = False,
        pass_manager: PassManager | None = None,
    ) -> None:

        if (not isinstance(sampler, BaseSamplerV1)) and (not isinstance(sampler, BaseSamplerV2)):
            raise ValueError(
                f"The sampler should be an instance of BaseSampler or BaseSamplerV2, "
                f"but got {type(sampler)}"
            )
        self._sampler: BaseSamplerV1 | BaseSamplerV2 = sampler
        self._pass_manager = pass_manager
        self._local = local
        self._default_options = Options()
        if options is not None:
            self._default_options.update_options(**options)
        BaseStateFidelity.__init__(self)  # pylint: disable=non-parent-init-called


class TrainQSVCKernel(Algorithm):
    """
    Train a quantum kernel with additional parameters to be optimized.
    The kernel will be optimized to provide maximum accuracy with a support vector classifier on the provided dataset.
    If no trainable parameters are provided, the algorithm will return
        a :class:`qiskit_machine_learning.kernels.FidelityQuantumKernel` kernel with a sampler assigned to the provided backend.
        Otherwise an instance of :class:`qiskit_machine_learning.kernels.TrainableFidelityQuantumKernel` with optimal 
        parameters and a sampler assigned to the provided backend will be returned.

    Args:
        kernel_circuit(QuantumCircuit): A parametrizable quantum circuit. The measurements will be used to produce kernel output.
        trainable_params(Sequence[Parameter] | None, optional): 
            The parameters to be optimized during training. If None no optimization will be done. Defaults to None.
    """

    _algorithm_format = 'tabular_ml'

    def __init__(
        self,
        kernel_circuit: QuantumCircuit,
        trainable_params: Sequence[Parameter] | None = None,
        **alg_kwargs
    ) -> None:
        super().__init__(**alg_kwargs)
        self.kernel = kernel_circuit
        self.trainable = trainable_params if trainable_params is not None else []

    def run(self, problem: Problem, backend: Backend, formatter: Callable[..., Any]) -> Result:
        X, y = formatter(problem)

        if not isinstance(X, np.ndarray):
            raise ValueError(f"X is not of type np.ndarray: received {type(X)}")

        if not isinstance(y, np.ndarray):
            raise ValueError(f"y is not of type np.ndarray: received {type(y)}")

        if isinstance(backend, (QiskitBackend, CirqBackend)):
            sampler = backend.samplerV1
        else:
            raise ValueError(f"The accepted backends are QiskitBackend and CirqBackend, got {type(backend)}")

        if len(self.trainable) == 0:
            return Result(
                best_bitstring='',
                best_energy=1,
                most_common_bitstring='',
                most_common_bitstring_energy=0,
                distribution={},
                energies={},
                num_of_samples=0,
                average_energy=0,
                energy_std=0,
                result=FidelityQuantumKernel(feature_map=self.kernel, fidelity=ComputeUncomputeCustom(sampler=sampler))
            )
        trainable_kernel = TrainableFidelityQuantumKernel(
            feature_map=self.kernel,
            fidelity=ComputeUncomputeCustom(sampler=sampler),
            training_parameters=self.trainable
        )
        kernel_trainer = QuantumKernelTrainer(trainable_kernel)
        kernel_trainer.fit(X, y)

        return Result(
            best_bitstring='',
            best_energy=1,
            most_common_bitstring='',
            most_common_bitstring_energy=0,
            distribution={},
            energies={},
            num_of_samples=0,
            average_energy=0,
            energy_std=0,
            result=kernel_trainer.quantum_kernel
        )



from collections import defaultdict
from collections.abc import Callable
from typing import Any
from inspect import signature
import warnings
import networkx as nx
from .base import Problem

__QL_ADAPTERS: dict[str, dict[str, Callable]] = defaultdict(lambda: {})  # adapters[to][from]
__QL_FORMATTERS: dict[type[Problem] | None, dict[str, Callable]] = defaultdict(lambda: {})  # formatters[problem][format]


def _get_callable_name(c: Callable) -> str:
    """
    Gets name of input callable, made it because callable classes don't have a __name__
    """
    try:
        return c.__name__
    except AttributeError:
        return str(c.__class__)


def _merge_dicts(dicts: list[dict]) -> dict:
    out = {}
    for d in dicts:
        out = out | d
    return out


class ProblemFormatter:
    """
    Converts input problem to a given format (input and output types determined by formatter and adapters in __init__)

    Probably shouldn't be constructed directly, call :py:func:`get_formatter()`
    """

    def __init__(self, formatter: Callable, adapters: list[Callable] | None = None):
        self.formatter = formatter
        self.adapters = [a['func'] for a in adapters] if adapters is not None else []
        self.adapter_requirements = _merge_dicts([a['formatter_requirements'] for a in adapters] if adapters is not None else [])

        self.formatter_sig = signature(self.formatter)

        self.run_params = {}

    def _formatter_call(self, run_params, *args, **kwargs):
        params_used = {k: v for k, v in run_params.items() if k in self.formatter_sig.parameters}

        unused_count = len(run_params) - len(params_used)
        if unused_count > 0:
            warnings.warn(
                f"{unused_count} unused parameters. {_get_callable_name(self.formatter)} does not accept {[k for k in run_params if not k in params_used]}", Warning)
        return self.formatter(*args, **kwargs, **params_used)

    def __call__(self, *args, **kwargs):
        # Reset bound params
        curr_run_params = dict(self.run_params)
        self.run_params = {}

        common_params = set(curr_run_params.keys()).intersection(set(self.adapter_requirements.keys()))
        if len(common_params) > 0:
            warnings.warn(
                f"Attempting to reassign parameter values required by one of the adapters: {common_params}, those params will not be set."
            )

        curr_run_params = curr_run_params | self.adapter_requirements

        out = self._formatter_call(curr_run_params, *args, **kwargs)
        for a in self.adapters:
            out = a(out)

        return out

    def get_pipeline(self) -> str:
        """
        Returns:
            String representing the conversion process: problem -> formatter -> adapters (if applicable)
        """
        return " -> ".join(
            [str(list(self.formatter_sig.parameters.keys())[0])] +
            [_get_callable_name(self.formatter)] +
            [_get_callable_name(fn) for fn in self.adapters]
        )

    def set_run_param(self, param: str, value: Any) -> None:
        """
        Sets a parameter to be used during next conversion.

        Args:
            param (str): parameter key
            value (str): parameter value
        """
        self.run_params[param] = value

    def set_run_params(self, params: dict[str, Any]) -> None:
        """
        Sets multiple parameters to be used during next conversion.

        Args:
            params (dict[str, Any]): parameters to be set
        """
        for k, v in params.items():
            self.set_run_param(k, v)


def adapter(translates_from: str, translates_to: str, **kwargs) -> Callable:
    """
    Register a function as an adapter from one problem format to another.

    Args:
        translates_from (str): Input format
        translates_to (str): Output format

    Returns:
        Same function
    """
    def decorator(func):
        if isinstance(func, type):
            func = func()
        __QL_ADAPTERS[translates_to][translates_from] = {'func': func, 'formatter_requirements': kwargs}
        return func
    return decorator


def formatter(problem: type[Problem] | None, alg_format: str):
    """
    Register a function as a formatter for a given problem type to a given format.

    Args:
        problem (type[Problem]): Input problem type
        alg_format (str): Output format

    Returns:
        Same function
    """
    def decorator(func):
        if isinstance(func, type):
            func = func()
        __QL_FORMATTERS[problem][alg_format] = func
        return func
    return decorator


def _find_shortest_adapter_path(problem: type[Problem], alg_format: str) -> list[str] | None:
    """
    Creates directed graph of possible conversions between formats and finds shortest path of formats between problem and alg_format.

    Returns:
        List of formats or None if no path was found.
    """
    G = nx.DiGraph()
    for problem_node in __QL_FORMATTERS[problem]:
        G.add_edge("__problem__", problem_node)

    for out_form in __QL_ADAPTERS:
        for in_form in __QL_ADAPTERS[out_form]:
            G.add_edge(in_form, out_form)

    if not G.has_node(alg_format):
        return None

    path = nx.shortest_path(G, "__problem__", alg_format)
    assert isinstance(path, list) or path is None, "Something went wrong in `nx.shortest_path`"
    return path


def get_formatter(problem: type[Problem], alg_format: str) -> ProblemFormatter:
    """
    Creates a ProblemFormatter that converts a given Problem subclass into the requested format.

    Args:
        problem (type[Problem]): Input problem type
        alg_format (str): Desired output format

    Returns:
        ProblemFormatter meeting the desired criteria.

    Raises:
        ValueError: If no combination of adapters can achieve conversion from problem to desired format.
    """
    formatter, adapters = None, None
    available_problem_formats = set(__QL_FORMATTERS[problem].keys())
    if alg_format in available_problem_formats:
        formatter = __QL_FORMATTERS[problem][alg_format]
    else:
        path = _find_shortest_adapter_path(problem, alg_format)

        if path is None:
            formatter = default_formatter
        else:
            formatter = __QL_FORMATTERS[problem][path[1]]
            adapters = []
            for i in range(1, len(path)-1):
                adapters.append(__QL_ADAPTERS[path[i+1]][path[i]])

    return ProblemFormatter(formatter, adapters)


@formatter(None, 'none')
def default_formatter(problem: Problem):
    return problem.instance



""" File with templates """
from typing import Literal
import json
import pickle
import logging
from qlauncher.base.adapter_structure import get_formatter, ProblemFormatter
from qlauncher.base import Problem, Algorithm, Backend, Result
from qlauncher.problems import Raw


class QLauncher:
    """
    QLauncher class.

    Qlauncher is used to run quantum algorithms on specific problem instances and backends.
    It provides methods for binding parameters, preparing the problem, running the algorithm, and processing the results.

    Attributes:
        problem (Problem): The problem instance to be solved.
        algorithm (Algorithm): The quantum algorithm to be executed.
        backend (Backend, optional): The backend to be used for execution. Defaults to None.
        path (str): The path to save the results. Defaults to 'results/'.
        binding_params (dict or None): The parameters to be bound to the problem and algorithm. Defaults to None.
        encoding_type (type): The encoding type to be used changing the class of the problem. Defaults to None.

    Example of usage::

            from qlauncher import QLauncher
            from qlauncher.problems import MaxCut
            from qlauncher.routines.qiskit import QAOA, QiskitBackend

            problem = MaxCut(instance_name='default')
            algorithm = QAOA()
            backend = QiskitBackend('local_simulator')

            launcher = QLauncher(problem, algorithm, backend)
            result = launcher.process(save_pickle=True)
            print(result)

    """

    def __init__(self, problem: Problem, algorithm: Algorithm, backend: Backend | None = None, logger: logging.Logger | None = None) -> None:

        if not isinstance(problem, Problem):
            problem = Raw(problem)

        self.problem: Problem = problem
        self.algorithm: Algorithm = algorithm
        self.backend: Backend | None = backend
        self.formatter: ProblemFormatter = get_formatter(self.problem._problem_id, self.algorithm._algorithm_format)

        if logger is None:
            logger = logging.getLogger('QLauncher')
        self.logger = logger

        self.result: Result | None = None

    def run(self, **kwargs) -> Result:
        """
        Finds proper formatter, and runs the algorithm on the problem with given backends.

        Returns:
            dict: The results of the algorithm execution.
        """

        self.formatter.set_run_params(kwargs)

        self.result = self.algorithm.run(self.problem, self.backend, formatter=self.formatter)
        self.logger.info('Algorithm ended successfully!')
        return self.result

    def save(self, path: str, save_format: Literal['pickle', 'txt', 'json'] = 'pickle'):
        """
        Save last run result to file

        Args:
            path (str): File path.
            save_format (Literal[&#39;pickle&#39;, &#39;txt&#39;, &#39;json&#39;], optional): Save format. Defaults to 'pickle'.

        Raises:
            ValueError: When no result is available or an incorrect save format was chosen
        """
        if self.result is None:
            raise ValueError("No result to save")

        # if not os.path.isfile(path):
        #     path = os.path.join(path, f'result-{datetime.now().isoformat(sep="_").replace(":","_")}.{save_format}')

        self.logger.info('Saving results to file: %s', str(path))
        if save_format == 'pickle':
            with open(path, mode='wb') as f:
                pickle.dump(self.result, f)
        elif save_format == 'json':
            with open(path, mode='w', encoding='utf-8') as f:
                json.dump(self.result.__dict__, f, default=fix_json)
        elif save_format == 'txt':
            with open(path, mode='w', encoding='utf-8') as f:
                f.write(str(self.result))
        else:
            raise ValueError(
                f'format: {save_format} in not supported try: pickle, txt, csv or json')


def fix_json(o: object):
    # if o.__class__.__name__ == 'SamplingVQEResult':
    #     parsed = self.algorithm.parse_samplingVQEResult(o, self._full_path)
    #     return parsed
    if o.__class__.__name__ == 'complex128':
        return repr(o)
    print(
        f'Name of object {o.__class__} not known, returning None as a json encodable')
    return None



def get_label(task, time):
    return f"{task.job}_{task.position},{time}"


class Task:
    def __init__(self, job, position, machine, duration):
        self.job = job
        self.position = position
        self.machine = machine
        self.duration = duration

    def __repr__(self):
        return ("{{job: {job}, position: {position}, machine: {machine}, duration:"
                " {duration}}}").format(**vars(self))


class KeyList:

    def __init__(self, array, key_function):
        self.array = array
        self.key_function = key_function

    def __len__(self):
        return len(self.array)

    def __getitem__(self, index):
        item = self.array[index]
        key = self.key_function(item)
        return key


class JobShopScheduler:
    def __init__(self, job_dict, max_time=None):
        self.tasks = []
        self.last_task_indices = []
        self.max_time = max_time
        self.H = 0
        self.H_vars = {}
        self.absurd_times = set()
        self._process_data(job_dict)

    def _process_data(self, jobs):
        tasks = []
        last_task_indices = [-1]
        total_time = 0

        for job_name, job_tasks in jobs.items():
            last_task_indices.append(last_task_indices[-1] + len(job_tasks))

            for i, (machine, time_span) in enumerate(job_tasks):
                tasks.append(Task(job_name, i, machine, time_span))
                total_time += time_span

        self.tasks = tasks
        self.last_task_indices = last_task_indices[1:]

        if self.max_time is None:
            self.max_time = total_time

    def _remove_absurd_times(self, disable_till: dict, disable_since, disabled_variables):
        predecessor_time = 0
        current_job = self.tasks[0].job
        for task in self.tasks:
            if task.job != current_job:
                predecessor_time = 0
                current_job = task.job

            for t in range(predecessor_time):
                label = get_label(task, t)
                self.absurd_times.add(label)

            predecessor_time += task.duration

        successor_time = -1
        current_job = self.tasks[-1].job
        for task in self.tasks[::-1]:
            if task.job != current_job:
                successor_time = -1
                current_job = task.job

            successor_time += task.duration
            for t in range(successor_time):
                label = get_label(task, (self.max_time - 1) - t)
                self.absurd_times.add(label)

        for task in self.tasks:
            if task.machine in disable_till.keys():
                for i in range(disable_till[task.machine]):
                    label = get_label(task, i)
                    self.absurd_times.add(label)
            elif task.machine in disable_since.keys():
                for i in range(disable_since[task.machine], self.max_time):
                    label = get_label(task, i)
                    self.absurd_times.add(label)

        for label in disabled_variables:
            self.absurd_times.add(label)



""" IBM backend class for Qiskit routines """
from typing import Literal

from qiskit.providers import BackendV1, BackendV2
from qiskit_ibm_runtime import EstimatorV2, SamplerV2
from qiskit_ibm_runtime import Session, Options

from qlauncher.routines.qiskit.backends.qiskit_backend import QiskitBackend


class IBMBackend(QiskitBackend):
    """ 
    An extension of QiskitBackend providing support for IBM sessions.

    Attributes:
        session (Session | None, optional): The session to use with name 'device'.
    """

    def __init__(
        self,
        name: Literal['local_simulator', 'backendv1v2', 'session'],
        options: Options | None = None,
        backendv1v2: BackendV1 | BackendV2 | None = None,
        auto_transpile_level: Literal[0, 1, 2, 3] | None = None,
        session: Session | None = None,
    ) -> None:
        self.session = session
        super().__init__(name, options, backendv1v2, auto_transpile_level)

    @property
    def setup(self) -> dict:
        return {
            'name': self.name,
            'session': self.session
        }

    def _set_primitives_on_backend_name(self) -> None:
        if self.name == 'local_simulator':
            super()._set_primitives_on_backend_name()
            return
        self._auto_assign = True
        if self.name == 'backendv1v2':
            if self.backendv1v2 is None:
                raise AttributeError(
                    'Please indicate a backend when in backendv1v2 mode.')
            self.estimator = EstimatorV2(self.backendv1v2)
            self.sampler = SamplerV2(self.backendv1v2)

        elif self.name == 'session':
            if self.session is None:
                raise AttributeError(
                    'Please indicate a session when in session mode.')
            else:
                self.estimator = EstimatorV2(mode=self.session, options=self.options)
                self.sampler = SamplerV2(mode=self.session, options=self.options)
        else:
            raise ValueError(
                " ".join([
                    f"Unsupported mode for this backend:'{self.name}'."
                    "Please use one of the following: ['local_simulator', 'backendv1v2', 'session']"
                ])
            )

        self._configure_auto_behavior()



import cProfile
import pstats
from qlauncher.base import Algorithm, Backend, Problem, Result
from .qlauncher import QLauncher


class ProfilingLauncher(QLauncher):
    """ Launcher made for debugging purposes of algorithms and other launchers focusing on performance issues 

    Attributes:
        profiler_path (str) path where to save the profiling results.

    """

    def __init__(self, problem: Problem, algorithm: Algorithm, backend: Backend, profiler_path: str = 'profiling-results.prof'):
        super().__init__(problem, algorithm, backend)
        self._profiler_path = profiler_path

    def run(self) -> Result:
        with cProfile.Profile() as pr:
            result = super().run()
        stats = pstats.Stats(pr)
        stats.sort_stats(pstats.SortKey.TIME)
        stats.reverse_order()
        stats.print_stats()
        stats.dump_stats(self._profiler_path)
        return result



""" Base backend class for Qiskit routines. """
from typing import Literal

from qiskit.providers import BackendV1, BackendV2
from qiskit.primitives import (
    BackendSamplerV2,
    BackendEstimatorV2,
    StatevectorEstimator,
    StatevectorSampler,
    Sampler
)

from qiskit_ibm_runtime import Options

from qlauncher.base import Backend
from qlauncher.routines.qiskit.adapters import SamplerV2ToSamplerV1Adapter

from qlauncher.routines.qiskit.backends.utils import (
    set_estimator_auto_run_behavior, set_sampler_auto_run_behavior,
    AUTO_TRANSPILE_ESTIMATOR_TYPE, AUTO_TRANSPILE_SAMPLER_TYPE
)


class QiskitBackend(Backend):
    """
    Base class for backends compatible with qiskit.

    Attributes:
        name (str): The name of the backend.
        options (Options | None, optional): The options for the backend. Defaults to None.
        backendv1v2 (BackendV1 | BackendV2 | None, optional): Predefined backend to use with name 'backendv1v2'. Defaults to None.
        sampler (BaseSamplerV2): The sampler used for sampling.
        estimator (BaseEstimatorV2): The estimator used for estimation.
    """

    def __init__(
        self,
        name: Literal['local_simulator', 'backendv1v2'] | str,
        options: Options | None = None,
        backendv1v2: BackendV1 | BackendV2 | None = None,
        auto_transpile_level: Literal[0, 1, 2, 3] | None = None
    ) -> None:
        """
        Args:
            **name (Literal[&#39;local_simulator&#39;, &#39;backendv1v2&#39;] | str)**: Name or mode of operation,
            'backendv1v2' allows for using a specific backend simulator.
            **options (Options | None, optional)**: Defaults to None.
            **backendv1v2 (BackendV1 | BackendV2 | None, optional)**: 
                Used with name 'backendv1v2', sampler and estimator will use it. Defaults to None.
            **auto_transpile_level (Literal[0, 1, 2, 3] | None, optional)**: 
                Optimization level for automatic transpilation of circuits.  
            - None: Don't transpile.
            - 0: No optimization (only transpile to compatible gates).
            - 1: Light optimization.
            - 2: Heavy optimization.
            - 3: Heaviest optimization.
            Defaults to None.
        """
        super().__init__(name)
        self.options = options
        self.backendv1v2 = backendv1v2
        self._auto_transpile_level = auto_transpile_level
        self._auto_assign = False
        self._samplerV1: Sampler | None = None
        self._set_primitives_on_backend_name()

    @property
    def samplerV1(self) -> Sampler:
        if self._samplerV1 is None:
            self._samplerV1 = SamplerV2ToSamplerV1Adapter(self.sampler)
        return self._samplerV1

    def _set_primitives_on_backend_name(self):
        if self.name == 'local_simulator':
            self.estimator = StatevectorEstimator()
            self.sampler = StatevectorSampler()
        elif self.name == 'backendv1v2':
            if self.backendv1v2 is None:
                raise AttributeError(
                    'Please indicate a backend when in backendv1v2 mode.')
            self.estimator = BackendEstimatorV2(backend=self.backendv1v2)
            self.sampler = BackendSamplerV2(backend=self.backendv1v2)

        else:
            raise ValueError(f"Unsupported mode for this backend:'{self.name}'")

        self._configure_auto_behavior()

    def _configure_auto_behavior(self):
        """
        Set auto transpilation and/or auto assignment if turned on, on estimator and sampler if compatible.
        """
        do_transpile, level = self._auto_transpile_level != None, int(
            self._auto_transpile_level if self._auto_transpile_level != None else 0)
        if isinstance(self.estimator, AUTO_TRANSPILE_ESTIMATOR_TYPE.__constraints__):
            self.estimator = set_estimator_auto_run_behavior(
                self.estimator,
                auto_transpile=do_transpile,
                auto_transpile_level=level,
                auto_assign=self._auto_assign
            )
        if isinstance(self.sampler, AUTO_TRANSPILE_SAMPLER_TYPE.__constraints__):
            self.sampler = set_sampler_auto_run_behavior(
                self.sampler,
                auto_transpile=do_transpile,
                auto_transpile_level=level,
                auto_assign=self._auto_assign
            )
